# preprocessor.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


import pandas as pd
import os

# all datasets need:
# 1. source
# 2. destination
# 3. distance
# 4. source lat
# 5. source lng
# 6. destination lat
# 7. destination lng


class preprocessor:

    @staticmethod
    def _save(
        df: pd.DataFrame,
        targetType: str = "parquet"
    ) -> None:
        """
        Save the DataFrame to a file in the data directory.

        Parameters:
            df (pd.DataFrame): The DataFrame to be saved.
            targetType (str): The type of file to be saved. Defaults to "parquet".
        """
        from pathlib import Path

        # create data directory if it doesn't exist (should be on te same level as this parent folder)
        data_dir = Path(__file__).parent.parent / "data"
        data_dir.mkdir(parents=True, exist_ok=True)
        file_path = data_dir / f"fullDataset.{targetType}"

        if targetType == "csv":
            # Save the DataFrame to a csv file
            df.to_csv(file_path, index=False)
        else:
            # Save the DataFrame to a parquet file
            df.to_parquet(file_path, engine="pyarrow")

    @staticmethod
    def preprocess(
        path: str,
        sourceKey: str = "source",
        sourceNameKey: str = "source_name",
        destinationKey: str = "destination",
        destinationNameKey: str = "destination_name",
        distanceKey: str = "distance",
        sourceLatKey: str = "source_lat",
        sourceLngKey: str = "source_lng",
        destinationLatKey: str = "destination_lat",
        destinationLngKey: str = "destination_lng",
        targetType: str = "parquet"
    ) -> pd.DataFrame:
        """
        Preprocess a dataset by renaming columns to the desired format,
        calculating distances and adding the result to the dataframe.

        Parameters:
        path (str): path to the dataset
        sourceKey (str): key for the source column (default: "source")
        destinationKey (str): key for the destination column (default: "destination")
        distanceKey (str): key for the distance column (default: "distance")
        sourceLatKey (str): key for the source latitude column (default: "source_lat")
        sourceLngKey (str): key for the source longitude column (default: "source_lng")
        destinationLatKey (str): key for the destination latitude column (default: "destination_lat")
        destinationLngKey (str): key for the destination longitude column (default: "destination_lng")

        Returns:
        pd.DataFrame: the preprocessed dataframe
        """

        # check if file exists and read it into a df
        _, fType = os.path.splitext(path)
        if fType == ".csv":
            df = pd.read_csv(path)
        elif fType == ".parquet":
            df = pd.read_parquet(path)

        # get all column names
        cols = list(df.columns)

        # check if all required columns are present
        if any([
            sourceKey not in cols,
            sourceNameKey not in cols,
            destinationKey not in cols,
            destinationNameKey not in cols,
            sourceLatKey not in cols,
            sourceLngKey not in cols,
            destinationLatKey not in cols,
            destinationLngKey not in cols
        ]):
            raise Exception("Invalid dataset")

        # rename columns to the desired format
        df.rename(columns={
            sourceKey: "source",
            sourceNameKey: "source_name",
            destinationKey: "destination",
            destinationNameKey: "destination_name",
            sourceLatKey: "source_lat",
            sourceLngKey: "source_lng",
            destinationLatKey: "destination_lat",
            destinationLngKey: "destination_lng",
            **({distanceKey: "distance"} if distanceKey in cols else {})
        }, inplace=True)

        # distance is already present return here
        if distanceKey in cols:
            preprocessor._save(df, targetType=targetType)
            return df[[
                "source",
                "source_name",
                "destination",
                "destination_name",
                "distance",
                "source_lat",
                "source_lng",
                "destination_lat",
                "destination_lng"
            ]]

        # calculate distance
        df["distance"] = preprocessor.haversine(df)

        # save df
        preprocessor._save(df, targetType=targetType)
        # return processed df
        return df[[
            "source",
            "source_name"
            "destination",
            "destination_name",
            "distance",
            "source_lat",
            "source_lng",
            "destination_lat",
            "destination_lng"
        ]]

    @staticmethod
    def haversine(df: pd.DataFrame) -> float:
        # use torch for vector calculation
        import torch
        # set device
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # compute vectorized haversine
        with torch.no_grad():
            # convert to radians
            lat1 = torch.deg2rad(torch.tensor(df["source_lat"].values, device=device))
            lng1 = torch.deg2rad(torch.tensor(df["source_lng"].values, device=device))
            lat2 = torch.deg2rad(torch.tensor(df["destination_lat"].values, device=device))
            lng2 = torch.deg2rad(torch.tensor(df["destination_lng"].values, device=device))

            # compute delta lat and delta lng
            dlat = lat2 - lat1
            dlng = lng2 - lng1
            # compute haversine
            a = torch.sin(dlat / 2)**2 + torch.cos(lat1) * torch.cos(lat2) * torch.sin(dlng / 2)**2
            c = 2 * torch.atan2(torch.sqrt(a), torch.sqrt(1 - a))

            distances = 6371 * c

            return distances.cpu().numpy()

    @staticmethod
    def combine(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
        # Combine the two DataFrames
        combined_df = pd.concat([df1, df2], axis=0)
        return combined_df



# dataclasses.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from ..graph import RouteGraph
import plotly.graph_objects as go


class GraphDisplay():

    def __init__(self, graph: RouteGraph, name: str = "Graph", iconSize: int = 10) -> None:
        self.graph: RouteGraph = graph
        self.name: str = name
        self.iconSize: int = iconSize

    def _toPlotlyFormat(
        self,
        nodeTransform=None,
        edgeTransform=None
    ):
        """
        transform the graph data into plotly format.to use the display function

        args:
            - nodeTransform: function to transform the node coordinates (default = None)
            - edgeTransform: function to transform the edge coordinates (default = None)
        returns:
            - None (modifies self.nodes and self.edges)
        """
        self.nodes = {
            f"{hub.hubType}-{hub.id}": {
                "coords": hub.coords,
                "hubType": hub.hubType,
                "id": hub.id
            }
            for hub in self.graph._allHubs()
        }

        self.edges = [
            {
                "from": f"{hub.hubType}-{hub.id}",
                "to": f"{self.graph.getHubById(dest).hubType}-{dest}",
                **edge.allMetrics
            }
            for hub in self.graph._allHubs()
            for _, edge in hub.outgoing.items()
            for dest, edge in edge.items()
        ]
        self.dim = max(len(node.get("coords")) for node in self.nodes.values())

        if nodeTransform is not None:
            expandedCoords = [node.get("coords") + [0] * (self.dim - len(node.get("coords"))) for node in self.nodes.values()]
            transformedCoords = nodeTransform(expandedCoords)
            for node, coords in zip(self.nodes.values(), transformedCoords):
                node["coords"] = coords

            self.dim = max(len(node.get("coords")) for node in self.nodes.values())

        if edgeTransform is not None:
            starts = [edge["from"] for edge in self.edges]
            startCoords = [self.nodes[start]["coords"] for start in starts]
            ends = [edge["to"] for edge in self.edges]
            endCoords = [self.nodes[end]["coords"] for end in ends]

            transformedEdges = edgeTransform(startCoords, endCoords)
            for edge, transformedEdge in zip(self.edges, transformedEdges):
                edge["curve"] = transformedEdge

    def display(
        self,
        nodeTransform=None,
        edgeTransform=None,
        displayEarth=False
    ):
        """
        function to display any 2D or 3D RouteGraph

        args:
            - nodeTransform: function to transform the node coordinates (default = None)
            - edgeTransform: function to transform the edge coordinates (default = None)
            - displayEarth: whether to display the earth as a background (default = False, only in 3D)

        returns:
            - None (modifies self.nodes and self.edges opens the plot in a browser)

        """
        # transform the graph
        self._toPlotlyFormat(nodeTransform, edgeTransform)
        # init plotly placeholders
        node_x, node_y, node_z, text, colors = [], [], [], [], []
        edge_x, edge_y, edge_z, edge_text = [], [], [], []

        # add all the nodes
        for node_key, node_data in self.nodes.items():
            x, y, *rest = node_data["coords"]
            node_x.append(x)
            node_y.append(y)
            if self.dim == 3:
                node_z.append(node_data["coords"][2])
            text.append(f"{node_data['id']}<br>Type: {node_data['hubType']}")
            colors.append(hash(node_data['hubType']) % 10)

        # add all the edges
        for edge in self.edges:
            # check if edge has been transformed
            if "curve" in edge:
                curve = edge["curve"]
                # add all the points of the edge
                for point in curve:
                    edge_x.append(point[0])
                    edge_y.append(point[1])
                    if self.dim == 3:
                        edge_z.append(point[2])
                edge_x.append(None)
                edge_y.append(None)
                # if 3d add the extra none to close the edge
                if self.dim == 3:
                    edge_z.append(None)
            else:
                source = self.nodes[edge["from"]]["coords"]
                target = self.nodes[edge["to"]]["coords"]

                edge_x += [source[0], target[0], None]
                edge_y += [source[1], target[1], None]

                if self.dim == 3:
                    edge_z += [source[2], target[2], None]

            # add text and hover display
            hover = f"{edge['from']} → {edge['to']}"
            metrics = {k: v for k, v in edge.items() if k not in ("from", "to", "curve")}
            if metrics:
                hover += "<br>" + "<br>".join(f"{k}: {v}" for k, v in metrics.items())
            edge_text.append(hover)

        if self.dim == 2:
            # ceate the plot in 2d
            node_trace = go.Scatter(
                x=node_x,
                y=node_y,
                mode="markers",
                hoverinfo="text",
                text=text,
                marker=dict(
                    size=self.iconSize,
                    color=colors,
                    colorscale="Viridis",
                    showscale=True
                )
            )

            edge_trace = go.Scatter(
                x=edge_x,
                y=edge_y,
                line=dict(width=2, color="#888"),
                hoverinfo="text",
                text=edge_text,
                mode="lines"
            )

        elif self.dim == 3:
            # create the plot in 3d
            node_trace = go.Scatter3d(
                x=node_x,
                y=node_y,
                z=node_z,
                mode="markers",
                hoverinfo="text",
                text=text,
                marker=dict(
                    size=self.iconSize,
                    color=colors,
                    colorscale="Viridis",
                    showscale=True
                )
            )

            edge_trace = go.Scatter3d(
                x=edge_x,
                y=edge_y,
                z=edge_z,
                line=dict(width=1, color="#888"),
                hoverinfo="text",
                text=edge_text,
                mode="lines",
                opacity=0.6
            )

        # create the plotly figure
        fig = go.Figure(data=[edge_trace, node_trace])
        # render earth / sphere in 3d
        if self.dim == 3 and displayEarth:
            try:
                import numpy as np
                R = 6369.9  # sphere radius
                u = np.linspace(0, 2 * np.pi, 50)   # azimuthal angle
                v = np.linspace(0, np.pi, 50)       # polar angle
                u, v = np.meshgrid(u, v)

                # Cartesian coordinates
                x = R * np.cos(u) * np.sin(v)
                y = R * np.sin(u) * np.sin(v)
                z = R * np.cos(v)
            except ImportError:
                raise ImportError("numpy is required to display the earth")

            sphere_surface = go.Surface(
                x=x, y=y, z=z,
                colorscale='Blues',
                opacity=1,
                showscale=False,
                hoverinfo='skip'
            )

            fig.add_trace(sphere_surface)

        fig.update_layout(title="Interactive Graph", showlegend=False, hovermode="closest")
        fig.show()

    @staticmethod
    def degreesToCartesian3D(coords):
        try:
            import torch
            C = torch.tensor(coords)
            if C.dim() == 1:
                C = C.unsqueeze(0)
            R = 6371.0
            lat = torch.deg2rad(C[:, 0])
            lng = torch.deg2rad(C[:, 1])
            x = R * torch.cos(lat) * torch.cos(lng)
            y = R * torch.cos(lat) * torch.sin(lng)
            z = R * torch.sin(lat)
            return list(torch.stack((x, y, z), dim=1).numpy())
        except ImportError:
            import math
            R = 6371.0
            output = []
            for lat, lng in coords:
                lat = math.radians(lat)
                lng = math.radians(lng)
                x = R * math.cos(lat) * math.cos(lng)
                y = R * math.cos(lat) * math.sin(lng)
                z = R * math.sin(lat)
                output.append([x, y, z])
            return output

    @staticmethod
    def curvedEdges(start, end, R=6371.0, H=0.05, n=20):
        try:
            # if torch and np are available calc vectorized graeter circle curves
            import numpy as np
            import torch

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

            start_np = np.array(start, dtype=np.float32)
            end_np = np.array(end, dtype=np.float32)

            start = torch.tensor(start_np, device=device)
            end = torch.tensor(end_np, device=device)
            start = start.float()
            end = end.float()

            # normalize to sphere
            start_norm = R * start / start.norm(dim=1, keepdim=True)
            end_norm = R * end / end.norm(dim=1, keepdim=True)

            # compute angle between vectors
            dot = (start_norm * end_norm).sum(dim=1, keepdim=True) / (R**2)
            dot = torch.clamp(dot, -1.0, 1.0)
            theta = torch.acos(dot).unsqueeze(2)  # shape: (num_edges,1,1)

            # linear interpolation along great circle
            t = torch.linspace(0, 1, n, device=device).view(1, n, 1)
            one_minus_t = 1 - t
            sin_theta = torch.sin(theta)
            sin_theta[sin_theta == 0] = 1e-6

            factor_start = torch.sin(one_minus_t * theta) / sin_theta
            factor_end = torch.sin(t * theta) / sin_theta

            curve = factor_start * start_norm.unsqueeze(1) + factor_end * end_norm.unsqueeze(1)

            # normalize to radius
            curve = R * curve / curve.norm(dim=2, keepdim=True)

            # apply radial lift at curve center using sin weight
            weight = torch.sin(torch.pi * t)  # 0 at endpoints, 1 at center
            curve = curve * (1 + H * weight)

            return curve
        except ImportError:
            # fallback to calculating quadratic bezier curves with math
            import math
            curves_all = []

            def multiply_vec(vec, factor):
                return [factor * x for x in vec]

            def add_vec(*vecs):
                return [sum(items) for items in zip(*vecs)]

            for startP, endP in zip(start, end):
                mid = [(s + e) / 2 for s, e in zip(startP, endP)]
                norm = math.sqrt(sum(c ** 2 for c in mid))
                mid_proj = [R * c / norm for c in mid]
                mid_arch = [c * (1 + H) for c in mid_proj]

                curve = []
                for i in range(n):
                    t_i = i / (n - 1)
                    one_minus_t = 1 - t_i
                    point = add_vec(
                        multiply_vec(startP, one_minus_t ** 2),
                        multiply_vec(mid_arch, 2 * one_minus_t * t_i),
                        multiply_vec(endP, t_i ** 2)
                    )
                    curve.append(point)

                curves_all.append(curve)

            return curves_all



# dataclasses.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from multimodalrouter import RouteGraph
from multimodalrouter.graphics import GraphDisplay
import os


# custom transform to make lat lng to x y (-> lng lat)
def NodeTransform(coords):
    for coord in coords:
        yield list((coord[0], coord[1]))


if __name__ == "__main__":
    path = os.path.dirname(os.path.abspath(__file__))
    # initialize the graph
    graph = RouteGraph(
        maxDistance=50,
        transportModes={"cell": "walk", },
        dataPaths={"cell": os.path.join(path, "data", "maze.csv")},
        compressed=False,
        drivingEnabled=False
    )

    graph.build()
    # init the display
    display = GraphDisplay(graph)
    # display the graph (uses the transform to swap lat lng to x y)
    display.display(nodeTransform=NodeTransform)



# dataclasses.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from dataclasses import dataclass
from enum import Enum
from abc import abstractmethod, ABC


class OptimizationMetric(Enum):
    DISTANCE = "distance"


class EdgeMetadata:
    __slots__ = ['transportMode', 'metrics']

    def __init__(self, transportMode: str = None, **metrics):
        self.transportMode = transportMode
        self.metrics = metrics  # e.g., {"distance": 12.3, "time": 15} NOTE distance is required by the graph

    def getMetric(self, metric: OptimizationMetric | str):
        if isinstance(metric, str):
            value = self.metrics.get(metric)
            if value is None:
                raise KeyError(f"Metric '{metric}' not found in EdgeMetadata")
            return value

        value = self.metrics.get(metric.value)
        if value is None:
            raise KeyError(f"Metric '{metric.value}' not found in EdgeMetadata")
        return value

    def copy(self):
        return EdgeMetadata(transportMode=self.transportMode, **self.metrics)

    @property
    def allMetrics(self):
        return self.metrics.copy()

    def __str__(self):
        return f"transportMode={self.transportMode}, metrics={self.metrics}"


class Hub:
    """Base hub class - using regular class instead of dataclass for __slots__ compatibility"""
    __slots__ = ['coords', 'id', 'outgoing', 'hubType']

    def __init__(self, coords: list[float], id: str, hubType: str):
        self.coords: list[float] = coords
        self.id = id
        self.hubType = hubType
        # dict like {mode -> {dest_id -> EdgeMetadata}}
        self.outgoing: dict[str, dict[str, EdgeMetadata]] = {}

    def addOutgoing(self, mode: str, dest_id: str, metrics: EdgeMetadata):
        if mode not in self.outgoing:
            self.outgoing[mode] = {}
        self.outgoing[mode][dest_id] = metrics

    def getMetrics(self, mode: str, dest_id: str) -> EdgeMetadata:
        return self.outgoing.get(mode, {}).get(dest_id, None)

    def getMetric(self, mode: str, dest_id: str, metric: str) -> float:
        connection = self.outgoing.get(mode, {}).get(dest_id)
        return getattr(connection, metric, None) if connection else None

    def clone(self) -> "Hub":
        new = Hub(self.coords[:], self.id, self.hubType)

        for mode, dests in self.outgoing.items():
            for dest_id, meta in dests.items():
                new.addOutgoing(mode, dest_id, meta.copy())

        return new

    def __hash__(self):
        return hash((self.hubType, self.id))


@dataclass
class Route:
    """Route class can use dataclass since it doesn't need __slots__"""
    path: list[tuple[str, str]]
    totalMetrics: EdgeMetadata
    optimizedMetric: OptimizationMetric

    @property
    def optimizedValue(self):
        return self.totalMetrics.getMetric(self.optimizedMetric)

    @property
    def flatPath(self, toStr=True):
        """Flatten the path into a list of hub IDs"""
        if not self.path:
            return []
        # get all source hubs plus the final destination
        path = [edge for edge in self.path]
        if not toStr:
            return path
        pathStr = ""
        for i, edge in enumerate(path):
            if i == 0:
                pathStr += f"Start: {edge[0]}"
                continue

            if len(edge) > 2 and isinstance(edge[2], EdgeMetadata):
                pathStr += f"\n\tEdge: ({str(edge[2])})\n-> {edge[0]}"
            else:
                pathStr += f"{edge[0]} -> {edge[1]}"
        return pathStr

    def asGraph(self, graph):
        """
        Creates a new RouteGraph with only the hubs in the route.
        It replicates the settings from the original graph.

        ### NOTE:
            * the graph in the argument should be the same as the graph from which the route was created.
            * hubs that are present in the route, but not found in the graph will be skipped

        :param:
            graph: the graph to replicate settings from

        :returns:
            a new RouteGraph with the specified settings and the added route
        """
        from . import RouteGraph
        subGraph = RouteGraph(
            maxDistance=graph.maxDrivingDistance,
            transportModes=graph.TransportModes,
            compressed=graph.compressed,
            extraMetricsKeys=graph.extraMetricsKeys,
            drivingEnabled=graph.drivingEnabled,
            sourceCoordKeys=graph.sourceCoordKeys,
            destCoordKeys=graph.destCoordKeys
        )
        # gets the hubs from the route (if not present in graph the hub will be dropped)
        hubs: list[Hub] = [graph.getHubById(edge[0]) for edge in self.path if graph.getHubById(edge[0])]

        copies = [hub.clone() for hub in hubs]

        # add all hubs to subGraph
        for hub in copies:
            subGraph.addHub(hub)

        # add links between consecutive hubs
        for prev, curr in zip(copies, copies[1:]):
            transpMode = graph.TransportModes[prev.hubType]

            meta = prev.getMetrics(transpMode, curr.id)
            if meta is None:
                # recompute distance
                distance = graph._hubToHubDistances([curr], [prev])[0][0].item()
                meta = EdgeMetadata(transportMode=transpMode, distance=distance)
            else:
                meta = meta.copy()

            subGraph._addLink(prev, curr, transpMode, **meta.metrics)

        return subGraph


@dataclass
class VerboseRoute(Route):
    """Uses base Route class but adds additional info to hold the edge metadata for every leg"""
    path: list[tuple[str, str, EdgeMetadata]]


class Filter(ABC):

    @abstractmethod
    def filterEdge(self, edge: EdgeMetadata) -> bool:
        """
        Return True if you want to keep the edge else False

        Args:
            edge (EdgeMetadata): Edge to filter

        Returns:
            bool: True if you want to keep the edge
        """
        pass

    @abstractmethod
    def filterHub(self, hub: Hub) -> bool:
        """
        Return True if you want to keep the hub else False

        Args:
            hub (Hub): Hub to filter

        Returns:
            bool: True if you want to keep the hub
        """
        pass

    def filter(self, start: Hub, end: Hub, edge: EdgeMetadata) -> bool:
        return self.filterHub(start) and self.filterHub(end) and self.filterEdge(edge)



# demo.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information

from multimodalrouter import RouteGraph
import os
import pandas as pd


def main():
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("matplotlib is not installed. Please install matplotlib to use this example.")

    path = os.path.dirname(os.path.abspath(__file__))
    # init the maze df for the plot
    mazeDf = pd.read_csv(os.path.join(path, "data", "maze.csv"))
    # init the plot
    plt.figure(figsize=(10, 10))
    # draw the maze
    # draw the maze (grid lines)
    for _, row in mazeDf.iterrows():
        plt.plot(
            [row.source_lng, row.destination_lng],   # x = "lng" column
            [row.source_lat, row.destination_lat],   # y = "lat" column
            "k-"
        )

    # initialize the graph
    graph = RouteGraph(
        maxDistance=50,
        transportModes={"cell": "walk", },
        dataPaths={"cell": os.path.join(path, "data", "maze.csv")},
        compressed=False,
        drivingEnabled=False
    )
    # build the graph
    graph.build()
    # find the shortest route
    route = graph.find_shortest_path(
        start_id="cell-(0, 0)",
        end_id="cell-(0, 9)",
        allowed_modes=["walk"],
        verbose=True,
        max_segments=100
    )
    # print the route
    print(route.flatPath if route else "No route found")
    # make the route blue in the plot
    s_prev = None
    for s, _, _ in route.path:
        if s_prev is not None:
            h1 = graph.getHubById(s_prev)
            h2 = graph.getHubById(s)
            # Swap coords so x=column, y=row
            plt.plot(
                [h1.coords[1], h2.coords[1]],  # x-axis
                [h1.coords[0], h2.coords[0]],  # y-axis
                "b-"
            )
        s_prev = s

    # display the plot
    plt.show()


if __name__ == "__main__":
    main()



# dataclasses.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from multimodalrouter import RouteGraph
from multimodalrouter.graphics import GraphDisplay
import os

if __name__ == "__main__":
    path = os.path.dirname(os.path.abspath(__file__))
    graph = RouteGraph(
        maxDistance=50,
        transportModes={"airport": "fly", },
        dataPaths={"airport": os.path.join(path, "data", "fullDataset.csv")},
        compressed=False,
    )

    graph.build()
    display = GraphDisplay(graph)
    display.display(
        displayEarth=True,
        nodeTransform=GraphDisplay.degreesToCartesian3D,
        edgeTransform=GraphDisplay.curvedEdges
    )



# graph.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from tqdm import tqdm
import zlib
import dill
import heapq
import os
import pandas as pd
from .dataclasses import Hub, EdgeMetadata, OptimizationMetric, Route, Filter, VerboseRoute
from threading import Lock
from collections import deque


class RouteGraph:

    def __init__(
        self,
        # sets the max distance to connect hubs with driving edges
        maxDistance: float,
        # dict like {hubtype -> "airport": "fly", "shippingport": "shipping"}
        transportModes: dict[str, str],
        # dict like {hubtype -> "airport": path -> "airports.csv", "shippingport": "shippingports.csv"}
        dataPaths: dict[str, str] = {},
        # if true model file will be compressed otherwise normal .dill file
        compressed: bool = False,
        # list of extra columns to add to the edge metadata (dynamically added to links when key is present in dataser)
        extraMetricsKeys: list[str] = [],
        # if true will connect hubs with driving edges
        drivingEnabled: bool = True,
        # a list of coordinate names for the source coords in the datasets (name to dataset matching is automatic)
        sourceCoordKeys: list[str] = ["source_lat", "source_lng"],
        # a list of coordinate names for the destination coords in the datasets (name to dataset matching is automatic)
        destCoordKeys: list[str] = ["destination_lat", "destination_lng"],
    ):
        self.sourceCoordKeys = set(sourceCoordKeys)
        self.destCoordKeys = set(destCoordKeys)

        self.compressed = compressed
        self.extraMetricsKeys = extraMetricsKeys
        self.drivingEnabled = drivingEnabled

        self.TransportModes = transportModes
        # hubtype -> {hubid -> Hub}
        self.Graph: dict[str, dict[str, Hub]] = {}

        # save the paths to the data in the state dict
        for key, value in dataPaths.items():
            setattr(self, key + "DataPath", value)
            self.Graph[key] = {}

        self.maxDrivingDistance = maxDistance

        self._lock = Lock()

    def __getstate__(self):
        state = self.__dict__.copy()

        # remove attributes that break pickle
        if "_lock" in state:
            del state["_lock"]

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # set the lock for thread safety
        from threading import Lock
        self._lock = Lock()

    # =========== public helpers ==========

    def findClosestHub(self, allowedHubTypes: list[str], coords: list[float]) -> Hub | None:
        """
        Find the closest hub of a given type to a given location.

        Args:
            hubType: Type of hub to find
            coords: list[float] = the coordinates of the location

        Returns:
            Hub instance if found, None otherwise
        """
        potentialHubs = []
        if allowedHubTypes is not None:
            for hubType in allowedHubTypes:
                potentialHubs.extend(self.Graph.get(hubType, {}).values())

        if allowedHubTypes is None:
            potentialHubs = list(self._allHubs())

        if not potentialHubs:
            return None

        tempHub = Hub(coords=coords, hubType="temp", id="temp")
        distances = self._hubToHubDistances([tempHub], potentialHubs).flatten()  # shape (n,)
        closest_hub = potentialHubs[distances.argmin()]
        return closest_hub

    def addHub(self, hub: Hub):
        """
        Add a hub to the graph. If the hub already exists, it will not be added a second time.

        Args:
            hub: Hub instance to add to the graph

        Returns:
            None
        """
        with self._lock:
            hubType = hub.hubType
            # if the hub type doesnt exist in the graph, create it first
            if hubType not in self.Graph:
                self.Graph[hubType] = {}
            # exit if the hub already exists
            if hub.id in self.Graph[hubType]:
                return
            # add the hub
            self.Graph[hubType][hub.id] = hub

    def getHub(self, hubType: str, id: str) -> Hub | None:
        """
        Get a hub from the graph by hub type and hub id.

        Args:
            hubType: Type of hub to get
            id: ID of the hub to get

        Returns:
            Hub instance if found, None otherwise
        """
        return self.Graph.get(hubType, {}).get(id)

    def getHubById(self, id: str) -> Hub | None:
        """
        Get a hub from the graph by its ID.

        Args:
            id: ID of the hub to get

        Returns:
            Hub instance if found, None otherwise
        """
        for hubType in self.Graph:
            hub = self.Graph[hubType].get(id)
            if hub:
                return hub
        return None

    def save(
        self,
        filepath: str = os.path.join(os.getcwd(), "..", "..", "..", "data"),
        compressed: bool = False
    ):
        """
        Save the RouteGraph to a file.

        Args:
            filepath (str): Path to save the graph to.
            saveMode (str, optional): Unused. Defaults to None.
            compressed (bool, optional): Whether to compress the saved graph. Defaults to False.

        Note: The graph is saved in the following format:
            - Compressed: <filepath>.zlib
            - Uncompressed: <filepath>.dill
        """
        with self._lock:

            # ensure correct compression type is set for loading the graph
            self.compressed = compressed
            os.makedirs(filepath, exist_ok=True)
            # save the graph
            pickled = dill.dumps(self)
            if compressed:
                compressed = zlib.compress(pickled)
                with open(os.path.join(filepath, "graph.zlib"), "wb") as f:
                    f.write(compressed)
            else:
                with open(os.path.join(filepath, "graph.dill"), "wb") as f:
                    f.write(pickled)

    @staticmethod
    def load(filepath: str, compressed: bool = False) -> "RouteGraph":
        """
        Load a RouteGraph from a file.

        Args:
            filepath (str): Path to load the graph from.
            compressed (bool, optional): Whether the saved graph is compressed. Defaults to False.

        Returns:
            RouteGraph: The loaded graph.
        """
        with open(filepath, "rb") as f:
            file_data = f.read()
            if compressed:
                decompressed = zlib.decompress(file_data)
                graph = dill.loads(decompressed)
            else:
                graph = dill.loads(file_data)
        return graph

    # ============= private helpers =============

    def _allHubs(self):
        for hubType in self.Graph:
            yield from self.Graph[hubType].values()

    def _addLink(
        self,
        hub1: Hub,
        hub2: Hub,
        mode: str,
        distance: float,
        bidirectional: bool = False,
        extraData: dict | None = None
    ):
        """
        Add a connection between two hubs, dynamically storing extra metrics.
        """
        with self._lock:
            if extraData is None:
                extraData = {}
            # combine required metrics with extra
            metrics = {"distance": distance, **extraData}
            edge = EdgeMetadata(transportMode=mode, **metrics)
            hub1.addOutgoing(mode, hub2.id, edge)
            if bidirectional:
                hub2.addOutgoing(mode, hub1.id, edge.copy())

    def _loadData(self, targetHubType: str):
        dataPath = getattr(self, targetHubType + "DataPath")
        if dataPath is None:
            raise ValueError(f"Data path for {targetHubType} is not set")
        fType = os.path.splitext(dataPath)[1]
        if fType == ".csv":
            data = pd.read_csv(dataPath)
        elif fType == ".parquet":
            data = pd.read_parquet(dataPath)
        else:
            raise ValueError(f"Unsupported file type {fType}")

        if data.empty:
            raise ValueError(f"{targetHubType} data is empty load the data first")

        return data

    def _generateHubs(self):
        """
        Generate Hub instances and link them with EdgeMetadata.
        Extra columns in the data will be added to EdgeMetadata dynamically.
        """
        # no lock needed since underlying methods have locks
        for hubType in self.Graph.keys():
            data = self._loadData(hubType)
            added = set()

            thisSourceKeys = self.sourceCoordKeys & set(data.columns)
            thisDestinationKeys = self.destCoordKeys & set(data.columns)

            # get required and extra columns
            required_cols = {
                "source", "destination",
                *thisSourceKeys,
                *thisDestinationKeys,
                "distance"
            }

            # collect extra data from the dataset columns that are not required but marked as extra
            extra_metric_cols = []
            for m in self.extraMetricsKeys:
                if m not in required_cols:
                    try:
                        extra_metric_cols.append(m)
                    except KeyError:
                        continue

            for row in tqdm(data.itertuples(index=False), desc=f"Generating {hubType} Hubs", unit="hub"):
                # create hubs if they don't exist
                if row.source not in added:
                    hub = Hub(coords=[getattr(row, k) for k in thisSourceKeys], id=row.source, hubType=hubType)
                    self.addHub(hub)
                    added.add(row.source)

                if row.destination not in added:
                    hub = Hub(coords=[getattr(row, k) for k in thisDestinationKeys], id=row.destination, hubType=hubType)
                    self.addHub(hub)
                    added.add(row.destination)

                # get extra metrics
                extra_metrics = {col: getattr(row, col) for col in extra_metric_cols}

                # link with the extra metrics
                self._addLink(
                    hub1=self.Graph[hubType][row.source],
                    hub2=self.Graph[hubType][row.destination],
                    mode=self.TransportModes[hubType],
                    distance=row.distance, # distance metric is absolutely required for all links
                    extraData=extra_metrics
                )

    def _hubToHubDistances(self, hub1: list[Hub], hub2: list[Hub]):
        """
        Compute full pairwise distance matrix between two lists of hubs using Haversine.

        Args:
            hub1: list of Hub
            hub2: list of Hub

        Returns:
            numpy.ndarray of shape (len(hub1), len(hub2))
        """
        import torch
        R = 6371.0
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        with torch.no_grad():
            lat1 = torch.deg2rad(torch.tensor([h.coords[0] for h in hub1], device=device))
            lng1 = torch.deg2rad(torch.tensor([h.coords[1] for h in hub1], device=device))
            lat2 = torch.deg2rad(torch.tensor([h.coords[0] for h in hub2], device=device))
            lng2 = torch.deg2rad(torch.tensor([h.coords[1] for h in hub2], device=device))

            lat1 = lat1.unsqueeze(1)
            lng1 = lng1.unsqueeze(1)
            lat2 = lat2.unsqueeze(0)
            lng2 = lng2.unsqueeze(0)

            dlat = lat2 - lat1
            dlng = lng2 - lng1

            a = torch.sin(dlat / 2) ** 2 + torch.cos(lat1) * torch.cos(lat2) * torch.sin(dlng / 2) ** 2
            c = 2 * torch.atan2(torch.sqrt(a), torch.sqrt(1 - a))
            distances = R * c

        return distances.cpu().numpy()

    # ============= public key functions =============

    def build(self):
        self._generateHubs()
        # exit here if not driving edges are allowed
        if not self.drivingEnabled:
            return

        # build driving edges
        hubTypes = list(self.Graph.keys())
        for i, hubType1 in enumerate(hubTypes):
            hubs1 = list(self.Graph[hubType1].values())
            for _, hubType2 in enumerate(hubTypes[i:], start=i):
                hubs2 = list(self.Graph[hubType2].values())
                distances = self._hubToHubDistances(hubs1, hubs2)

                for hi, hub1 in enumerate(hubs1):
                    for hj, hub2 in enumerate(hubs2):
                        if hub1.id == hub2.id:
                            continue

                        d = distances[hi, hj]
                        if d <= self.maxDrivingDistance:
                            self._addLink(
                                hub1=hub1,
                                hub2=hub2,
                                mode="car",  # explicitly set driving
                                distance=d,
                                bidirectional=True,
                                # no extra metrics for default drive nodes
                            )

    def find_shortest_path(
        self,
        start_id: str,
        end_id: str,
        allowed_modes: list[str] = None,
        optimization_metric: OptimizationMetric | str = OptimizationMetric.DISTANCE,
        max_segments: int = 10,
        verbose: bool = False,
        custom_filter: Filter = None,
    ) -> Route | VerboseRoute | None:
        """
        Find the optimal path between two hubs using Dijkstra

        Args:
            start_id: ID of the starting hub
            end_id: ID of the destination hub
            optimization_metric: Metric to optimize for (distance, time, cost, etc.) (must exist in EdgeMetadata)
            allowed_modes: List of allowed transport modes (default: all modes)
            max_segments: Maximum number of segments allowed in route

        Returns:
            Route object with the optimal path, or None if no path exists
        """

        # check if start and end hub exist
        start_hub = self.getHubById(start_id)
        end_hub = self.getHubById(end_id)

        if start_hub is None:
            raise ValueError(f"Start hub '{start_id}' not found in graph")
        if end_hub is None:
            raise ValueError(f"End hub '{end_id}' not found in graph")

        if allowed_modes is None:
            allowed_modes = list(self.TransportModes.values())

        if start_id == end_id:
            # create a route with only the start hub
            # no verbose since no edges are needed
            return Route(
                path=[(start_id, "")],
                totalMetrics=EdgeMetadata(),
                optimizedMetric=optimization_metric,
            )

        if verbose:
            # priority queue: (metric_value, hub_id, path_with_modes, accumulated_metrics)
            pq = [(0.0, start_id, [(start_id, "", EdgeMetadata())], EdgeMetadata())]
        else:
            # priority queue: (metric_value, hub_id, path_with_modes, accumulated_metrics)
            pq = [(0.0, start_id, [(start_id, "")], EdgeMetadata())]

        visited = {} # dict like {hub_id : metric_value}

        while pq:
            # get the current path data
            # optim metric,         hub id,       path with modes, accumulated metrics (edgeMetadata object)
            current_metric_value, current_hub_id, path_with_modes, accumulated_metrics = heapq.heappop(pq)

            # skip this if a better path exists
            if current_hub_id in visited and visited[current_hub_id] <= current_metric_value:
                continue
            # mark as visited
            visited[current_hub_id] = current_metric_value

            # check if this is the end hub
            if current_hub_id == end_id:
                if verbose:
                    return Route(
                        path=path_with_modes,
                        totalMetrics=accumulated_metrics,
                        optimizedMetric=optimization_metric,
                    )

                return Route(
                    path=path_with_modes,
                    totalMetrics=accumulated_metrics,
                    optimizedMetric=optimization_metric,
                )

            # skip if too many segments
            if len(path_with_modes) > max_segments:
                continue

            # get the current hub
            current_hub = self.getHubById(current_hub_id)
            if current_hub is None:
                continue

            # test all outgoing connections from the current hub
            for mode in allowed_modes: # iter over the allowed transport modes
                if mode in current_hub.outgoing: # check if the mode has outgoing connections
                    # iter over all outgoing links with the selected transport type
                    for next_hub_id, connection_metrics in current_hub.outgoing[mode].items():
                        if connection_metrics is None: # skip if the connection has no metrics
                            continue

                        try:
                            next_hub = self.getHubById(next_hub_id)
                        except KeyError:
                            raise ValueError(
                                f"Hub with ID '{next_hub_id}' not found in graph! But it is connected to hub '{current_hub_id}' via mode '{mode}'." # noqa: E501
                            )
                        if custom_filter is not None and not custom_filter.filter(current_hub, next_hub, connection_metrics):
                            continue

                        # get the selected metric alue for this connection
                        connection_value = connection_metrics.getMetric(optimization_metric)
                        new_metric_value = current_metric_value + connection_value

                        # skip if a better hub to get here exists
                        if next_hub_id in visited and visited[next_hub_id] <= new_metric_value:
                            continue

                        # create a new edge obj for the combined metrics  |  None bc modes my change between edges
                        new_accumulated_metrics = EdgeMetadata(transportMode=None, **accumulated_metrics.metrics)
                        # accumulate metrics
                        for metric_name, metric_value in connection_metrics.metrics.items():
                            if isinstance(metric_value, (int, float)):
                                new_accumulated_metrics.metrics[metric_name] = new_accumulated_metrics.metrics.get(metric_name, 0) + metric_value # noqa: E501
                            else:
                                # ignore non-numeric metrics for accumulation (maybe combine strings here)
                                new_accumulated_metrics.metrics[metric_name] = metric_value

                        # combine to form a new path
                        if verbose:
                            new_path = path_with_modes + [(next_hub_id, mode, connection_metrics)]
                        else:
                            new_path = path_with_modes + [(next_hub_id, mode)]
                        # push to the priority queue for future exploration
                        heapq.heappush(pq, (new_metric_value, next_hub_id, new_path, new_accumulated_metrics))

        return None

    def radial_search(
        self,
        hub_id: str,
        radius: float,
        optimization_metric: OptimizationMetric | str = OptimizationMetric.DISTANCE,
        allowed_modes: list[str] = None,
        custom_filter: Filter = None,
    ) -> list[float, Hub]:
        """
        Find all hubs within a given radius of a given hub
        (Note: distance is measured from the connecting paths not direct)

        Args:
            hub_id: ID of the center hub
            radius: maximum distance from the center hub
            optimization_metric: metric to optimize for (e.g. distance, time, cost)
            allowed_modes: list of allowed transport modes (default: None => all modes)

        Returns:
            list of tuples containing the metric value and the corresponding hub object
        """

        center = self.getHubById(hub_id)
        if center is None:
            return [center]

        if allowed_modes is None:
            allowed_modes = list(self.TransportModes.values())

        hubsToSearch = deque([center])
        queued = set([hub_id])
        reachableHubs: dict[str, tuple[float, Hub]] = {hub_id: (0.0, center)}

        while hubsToSearch:
            hub = hubsToSearch.popleft() # get the current hub to search
            currentMetricVal, _ = reachableHubs[hub.id] # get the current metric value
            for mode in allowed_modes:
                outgoing = hub.outgoing.get(mode, {}) # find all outgoing connections
                # dict like {dest_id: EdgeMetadata}
                for id, edgemetadata in outgoing.items(): # iter over outgoing connections
                    thisMetricVal = edgemetadata.getMetric(optimization_metric)
                    if thisMetricVal is None:
                        continue
                    nextMetricVal = currentMetricVal + thisMetricVal
                    if nextMetricVal > radius:
                        continue
                    knownMetric = reachableHubs.get(id, None)
                    destHub = self.getHubById(id)
                    if custom_filter is not None and not custom_filter.filter(hub, destHub, edgemetadata):
                        continue
                    # only save smaller metric values
                    if knownMetric is None or knownMetric[0] > nextMetricVal:
                        reachableHubs.update({id: (nextMetricVal, destHub)})
                    if id not in queued:
                        queued.add(id)
                        hubsToSearch.append(destHub)

        return [v for v in reachableHubs.values()]

    def compare_routes(
        self,
        start_id: str,
        end_id: str,
        allowed_modes: list[str],
        metrics_to_compare: list[OptimizationMetric] = None
    ) -> dict[OptimizationMetric, Route]:
        """
        Find optimal routes for different metrics and compare them

        Returns:
            Dictionary mapping each optimization metric to its optimal route
        """
        if metrics_to_compare is None:
            metrics_to_compare = list(OptimizationMetric)

        results = {}
        for metric in metrics_to_compare:
            route = self.find_shortest_path(start_id, end_id, optimization_metric=metric, allowed_modes=allowed_modes)
            if route:
                results[metric] = route

        return results



# router.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from ..graph import RouteGraph
import argparse
import os


def main():
    graph = RouteGraph.load(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "..", "data", "graph.dill"),
        compressed=False
    )

    parser = argparse.ArgumentParser(
        description="parse the arguments"
    )
    parser.add_argument(
        "--start",
        nargs="+",
        type=float,
        required=True,
        help="Start coordinates"
    )
    parser.add_argument(
        "--end",
        nargs="+",
        type=float,
        required=True,
        help="End coordinates"
    )
    parser.add_argument(
        "--allowedModes",
        nargs="+",
        type=str,
        default=["car"],
        help="Allowed transport modes"
    )
    parser.add_argument(
        "--maxSegments",
        type=int,
        default=10,
        help="Maximum number of segments in the route"
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Verbose output for the paths"
    )
    args = parser.parse_args()

    start_hub = graph.findClosestHub(["airport"], args.start)
    end_hub = graph.findClosestHub(["airport"], args.end)

    if start_hub is None or end_hub is None:
        print("One of the airports does not exist in the graph")
        return

    route = graph.find_shortest_path(start_id=start_hub.id,
                                     end_id=end_hub.id,
                                     allowed_modes=args.allowedModes,
                                     max_segments=args.maxSegments,
                                     verbose=args.verbose)

    print(route.flatPath if route else "No route found")


if __name__ == "__main__":
    main()



from .graph import RouteGraph, Hub, EdgeMetadata, OptimizationMetric, Route, VerboseRoute, Filter
from .utils import preprocessor

__all__ = ["RouteGraph", "Hub", "EdgeMetadata", "OptimizationMetric", "Route", "VerboseRoute", "preprocessor", "Filter"]



# demo.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information

from multimodalrouter import RouteGraph
import os


def main():
    path = os.path.dirname(os.path.abspath(__file__))
    # initialize the graph
    graph = RouteGraph(
        maxDistance=50,
        transportModes={"airport": "plane", },
        dataPaths={"airport": os.path.join(path, "data", "fullDataset.csv")},
        compressed=False,
    )
    # build the graph
    graph.build()
    # set start and end points
    start = [60.866699, -162.272996] # Atmautluak Airport
    end = [60.872747, -162.5247] # Kasigluk Airport

    start_hub = graph.findClosestHub(["airport"], start) # find the hubs
    end_hub = graph.findClosestHub(["airport"], end)
    # find the route
    route = graph.find_shortest_path(
        start_hub.id,
        end_hub.id,
        allowed_modes=["plane", "car"],
        verbose=True
    )
    # print the route
    print(route.flatPath if route else "No route found")


if __name__ == "__main__":
    main()



# build.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information.


from ..graph import RouteGraph
import argparse
import os


def main():
    print("Building graph...")
    parser = argparse.ArgumentParser(
        description="Collect key-value1-value2 triplets into two dicts"
    )
    parser.add_argument(
        "data",
        nargs="+",
        help="Arguments in groups of 3: hubType transportMode dataPath"
    )
    parser.add_argument(
        "--maxDist",
        type=int,
        default=50,
        help="Maximum distance to connect hubs with driving edges"
    )
    parser.add_argument(
        "--compressed",
        action="store_true",
        help="Whether to compress the saved graph (default: False)"
    )

    parser.add_argument(
        "--extraMetrics",
        nargs="+",
        default=[],
        help="Extra metrics to add to the edge metadata"
    )

    parser.add_argument(
        "--drivingEnabled",
        action="store_true",
        default=True,
        help="Whether to connect hubs with driving edges (default: True)"
    )
    path = os.path.dirname(os.path.abspath(__file__))
    parser.add_argument(
        "--Dir",
        type=str,
        default=os.path.join(path, "..", "..", "..", "data"),
        help="Directory to save the graph in (default: .)"
    )
    parser.add_argument(
        "--sourceKeys",
        nargs="+",
        default=["source_lat", "source_lng"],
        help="Source keys to search the source coordinates for (default: ['source_lat', 'source_lng'])"
    )
    parser.add_argument(
        "--destKeys",
        nargs="+",
        default=["destination_lat", "destination_lng"],
        help="Destination keys to search the destination coordinates for (default: ['destination_lat', 'destination_lng'])"
    )

    args = parser.parse_args()

    if len(args.data) % 3 != 0:
        parser.error("Arguments must be in groups of 3: hubType transportMode dataPath")

    transportModes = {}
    dataPaths = {}

    for i in range(0, len(args.data), 3):
        key, val1, val2 = args.data[i], args.data[i + 1], args.data[i + 2]
        transportModes[key] = val1
        dataPaths[key] = val2

    graph = RouteGraph(
        maxDistance=args.maxDist,
        transportModes=transportModes,
        dataPaths=dataPaths,
        compressed=args.compressed,
        extraMetricsKeys=args.extraMetrics,
        drivingEnabled=args.drivingEnabled,
        sourceCoordKeys=args.sourceKeys,
        destCoordKeys=args.destKeys
    )

    graph.build()
    graph.save(filepath=args.Dir, compressed=args.compressed)

    print("Graph built and saved.")


if __name__ == "__main__":
    main()



# demo.py
# Copyright (c) 2025 Tobias Karusseit
# Licensed under the MIT License. See LICENSE file in the project root for full license information

import random
import pandas as pd


# simple cell class for the maze
class Cell:
    def __init__(self, x, y):
        self.id = f"cell-{x, y}"
        self.x = x
        self.y = y
        self.visited = False
        self.connected = []


def main():
    # init a 10x10 maze
    mazeHeight = 10
    mazeWidth = 10
    # init the cells
    cells = [Cell(x, y) for x, y in [(x, y) for x in range(mazeWidth) for y in range(mazeHeight)]]
    cellStack = []
    # random choose one cell and init the stack
    random.shuffle(cells)
    cellStack.append(cells.pop())
    # build the maze
    while cellStack:
        currentCell = cellStack[-1] # peek at the current cell
        # get all not et visited neighbors
        neighbors = [
            c for c in cells if not c.visited and (
                (c.x == currentCell.x + 1 and c.y == currentCell.y) or
                (c.x == currentCell.x - 1 and c.y == currentCell.y) or
                (c.y == currentCell.y + 1 and c.x == currentCell.x) or
                (c.y == currentCell.y - 1 and c.x == currentCell.x)
            )
        ]

        if neighbors:
            # select a random neighbor
            random.shuffle(neighbors)
            neighbor = neighbors.pop()
            # connect the cells (remove walls)
            currentCell.connected.append(neighbor)
            neighbor.connected.append(currentCell)
            # mark the neighbor as visited
            neighbor.visited = True
            # add the neighbor to the stack
            cellStack.append(neighbor)
        else:
            # remove the current cell from the stack since no paths exist for it
            cellStack.pop()

    # init the dataframe
    data = pd.DataFrame(columns=[
        "source",
        "destination",
        "distance",
        "source_lat",
        "source_lng",
        "destination_lat",
        "destination_lng"
    ])
    # add the edges to the dataframe
    for cell in cells:
        for neighbor in cell.connected:
            data.loc[len(data)] = [cell.id, neighbor.id, 1, cell.x, cell.y, neighbor.x, neighbor.y]
    # save the dataframe
    data.to_csv("docs/examples/mazePathfinder/data/maze.csv", index=False)


if __name__ == "__main__":
    main()
